;;; x-pgp-sig.el -- Yet another PGP authenticating utility for Emacsen.

;;; Copyright (C) 1996,1997,1998 Katsumi Yamaoka
;;; Copyright (C) 1996,1997,1998 OKUNISHI Fujikazu
;;; Copyright (C) 1996,1997,1998 KOSEKI Yoshinori
;;; Copyright (C) 1996,1997,1998 Fumitoyo ONO
;;; Author: Katsumi Yamaoka   <yamaoka@jpl.org>
;;;         OKUNISHI Fujikazu <fuji0924@mbox.kyoto-inet.or.jp>
;;;         KOSEKI Yoshinori  <kose@yk.netlaputa.or.jp>
;;;         Fumitoyo ONO      <ono@sa.osk.sumikin.co.jp>
;;; Maintainer: Katsumi Yamaoka <yamaoka@jpl.org>
;;; Created:    98/01/09
;;; References: /anonymous@ftp.isc.org:/pub/pgpcontrol/README
;;;             ftp://ftp.isc.org/pub/pgpcontrol/README.html
;;; Keywords:   gnus, mh-e, mew, cmail, pgp, pgpverify, signcontrol, x-pgp-sig

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to the
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA 02111-1307, USA.

;;; Commands:
;;;  x-pgp-sig-set-signer
;;;     "Set PGP signer and passphrase."
;;;  x-pgp-sig-set-signer-maybe
;;;     "Set PGP signer and passphrase if necessary."
;;;  x-pgp-sig-switch-signing
;;;     "Judging whether to add X-PGP signature."
;;;  x-pgp-sig-verify
;;;     "Verify X-PGP signature in current message."

;;; Functions:
;;;  x-pgp-sig-sign
;;;     "Add X-PGP signature."
;;;  x-pgp-sig-mew-encode-coding-and-sign
;;;     "Encode the message to the network code and sign."
;;--------------------------------------------------------------------

;;; Code:

(or (boundp 'running-xemacs) (defconst running-xemacs (featurep 'xemacs)))

(defconst x-pgp-sig-version-number "1.2.10"
  "Version number for this version of x-pgp-sig.")
(defconst x-pgp-sig-version
  (format "Yet another PGP authenticating utility v%s"
	  x-pgp-sig-version-number)
  "Version string for this version of x-pgp-sig.")
(defconst x-pgp-sig-codename
  ;  "Helter Skelter"
  ;  "I've Got A Feeling"
  ;  "Jude Hey"
  ;  "Kome Together"
  ;  "Long Tall Sally"
  "Mother Nature's Son"
  ;  "Nowhere Man"
  ;  "Old Brown Shoe"
  ;  "Piggies"
  ;  "Revolution"
  ;  "Sgt.Pepper's"
  ;  "Twist & Shout"
  ;  "With A Little Help From My Friends"
  ;  "You Never Give Me Your Money"
  ;  "Dizzy Miss Lizzy"
  "Codename of this version of x-pgp-sig (a string).")

(require 'cl)

(unless (and (condition-case ()
		 (require 'custom)
	       (file-error nil))
	     (fboundp 'defgroup)
	     (fboundp 'defcustom))
  (require 'backquote)
  (defmacro defgroup (&rest args))
  (defmacro defcustom (symbol value &optional doc &rest args)
    (` (defvar (, symbol) (, value) (, doc))))
  )

; Silence the byte compiler.
(and
 (fboundp 'eval-when-compile)
 (eval-when-compile
   (save-excursion
     (beginning-of-defun)
     (eval-region (point-min) (point)))
   (let (case-fold-search)
     (if (string-match "alpha\\|beta" x-pgp-sig-version-number)
	 (mapcar
	  (function
	   (lambda (symbol)
	     (unless (boundp symbol)
	       (make-local-variable symbol)
	       (eval (list 'setq symbol nil)))))
	  '(:group
	    :prefix :type
	    cmail-current-folder
	    gnus-Article-buffer gnus-article-buffer gnus-article-display-hook
	    gnus-original-article-buffer
	    gnus-current-article gnus-local-domain gnus-summary-buffer
	    mail-host-address
	    mc-pgp-always-fetch
	    mew-cs-mime-trans mew-cs-noconv
	    mew-header-separator mew-summary-message-regex
	    mh-folder-filename
	    mime-editor/pgp-processing
	    to-ascii-display to-ascii-fileio to-ascii-process
	    to-kanji-display to-kanji-fileio to-kanji-process
	    ))
       (make-local-variable 'byte-compile-warnings)
       (setq byte-compile-warnings nil)
       ))))

(defgroup x-pgp-sig nil
  "Yet another PGP authenticating utility."
  :prefix "x-pgp-sig-"
  :group 'applications)

(defcustom x-pgp-sig-from
  (or
   (and (boundp 'user-mail-address)
	(stringp user-mail-address)
	user-mail-address)
   (and (boundp 'mail-host-address)
	(stringp mail-host-address)
	(concat (user-login-name) "@" mail-host-address))
   (and (boundp 'gnus-local-domain)
	(stringp gnus-local-domain)
	(concat (user-login-name) "@" gnus-local-domain))
   "nobody@nowhere")
  "*Signer's mail address."
  :group 'x-pgp-sig
  :type 'string)

(defcustom x-pgp-sig-default-signer
  (format "%s <%s>" (user-full-name) x-pgp-sig-from)
  "*PGP user ID."
  :group 'x-pgp-sig
  :type 'string)

(defcustom x-pgp-sig-pgp-program "pgp"
  "*PGP executable."
  :group 'x-pgp-sig
  :type 'string)

(defcustom x-pgp-sig-toMime-program "toMime"
  "*toMime executable."
  :group 'x-pgp-sig
  :type 'string)
(defcustom x-pgp-sig-toMime-options nil
  "*List of toMime options."
  :group 'x-pgp-sig
  :type 'list)

(defcustom x-pgp-sig-mew-encode-message-header-function
  'x-pgp-sig-mew-encode-message-header-with-toMime
  "*Function to call to encode the message of Mew. Legal values include
`x-pgp-sig-mew-encode-message-header-with-toMime' (the default),
`x-pgp-sig-mew-encode-message-header-with-semi' and
`x-pgp-sig-mew-encode-message-header-with-tm'."
  :group 'x-pgp-sig
  :type '(radio (function-item x-pgp-sig-mew-encode-message-header-with-toMime)
		(function-item x-pgp-sig-mew-encode-message-header-with-semi)
		(function-item x-pgp-sig-mew-encode-message-header-with-tm)
		function))

(defcustom x-pgp-sig-process-coding-system-for-output nil
  "*Coding system for PGP process (output from Mule/XEmacs)."
  :group 'x-pgp-sig
  :type 'coding-system)
(defcustom x-pgp-sig-process-kanji-code 2
  "*Kanji code for PGP process (NEmacs).
0:No-conversion will not work well."
  :group 'x-pgp-sig
  :type 'coding-system)

(defcustom x-pgp-sig-pgppath
  (or (getenv "PGPPATH") (expand-file-name "~/.pgp"))
  "*Environment variable \"PGPPATH\"."
  :group 'x-pgp-sig
  :type 'directory)

(defcustom x-pgp-sig-pubring
  (or (getenv "PUBRING")
      (concat (if (string-match "/+$" x-pgp-sig-pgppath)
		  (substring x-pgp-sig-pgppath 0 (match-beginning 0))
		x-pgp-sig-pgppath)
	      "/pubring.pgp"))
  "*Environment variable \"PUBRING\"."
  :group 'x-pgp-sig
  :type 'file)

(defcustom x-pgp-sig-secring
  (or (getenv "SECRING")
      (concat (if (string-match "/+$" x-pgp-sig-pgppath)
		  (substring x-pgp-sig-pgppath 0 (match-beginning 0))
		x-pgp-sig-pgppath)
	      "/secring.pgp"))
  "*Environment variable \"SECRING\"."
  :group 'x-pgp-sig
  :type 'file)

(defcustom x-pgp-sig-delete-last-empty-line-while-verifying t
  "*If non-nil, delete last empty line of message while verifying."
  :group 'x-pgp-sig
  :type 'boolean)

(defcustom x-pgp-sig-approved-field-body x-pgp-sig-from
  "*Address for Approved field."
  :group 'x-pgp-sig
  :type 'string)

(defcustom x-pgp-sig-x-info-field-body nil
  "*X-Info field body."
  :group 'x-pgp-sig
  :type 'string)

(defcustom x-pgp-sig-sign-fields
  '("Subject" "Control" "Message-ID" "Date" "From" "Sender"
    "X-Mailer" "X-Newsreader" "X-Face" "X-Face-Type")
  "*Fields to sign."
  :group 'x-pgp-sig
  :type 'list)

(defcustom x-pgp-sig-adjust-sign-fields-alist nil
  "*Adjust `x-pgp-sig-sign-fields' if matched text exists in narrowed header.

\((REGEXP1 (sub \"FIELD1\" \"FIELD2\" ...)
\          (add \"FIELD3\" \"FIELD4\" ...))
\ (REGEXP2 (sub \"FIELD5\" \"FIELD6\" ...)
\          (add \"FIELD7\" \"FIELD8\" ...)))

REGEXP can be a function as well."
  :group 'x-pgp-sig
  :type 'alist)

(defcustom x-pgp-sig-last-fields '("X-Face" "X-Face-Version" "X-Face-Type")
  "*PGP signed field will be added above these fields."
  :group 'x-pgp-sig
  :type 'list)

(defcustom x-pgp-sig-always-sign nil
  "*If non-nil, always add X-PGP signature."
  :group 'x-pgp-sig
  :type 'boolean)

(defcustom x-pgp-sig-add-signature-judging-function
  'x-pgp-sig-add-signature-y-or-n-p
  "*Function whether to add X-PGP signature."
  :group 'x-pgp-sig
  :type 'function)

(defcustom x-pgp-sig-load-hook nil
  "*Hook to be run after the x-pgp-sig package has been loaded."
  :group 'x-pgp-sig
  :type 'hook)

(defcustom x-pgp-sig-sign-hook nil
  "*Hook run after adding X-PGP-Sig field."
  :group 'x-pgp-sig
  :type 'hook)

(defcustom x-pgp-sig-prepare-sign-hook nil
  "*Hook run before signing pgp process."
  :group 'x-pgp-sig
  :type 'hook)

(defcustom x-pgp-sig-post-sign-hook nil
  "*Hook run after signing pgp process."
  :group 'x-pgp-sig
  :type 'hook)

(defcustom x-pgp-sig-good-signature-regexp "^Good signature from user.*$"
  "*Regular expression matching a PGP signature validation message."
  :group 'x-pgp-sig
  :type 'regexp)

(defcustom x-pgp-sig-bad-signature-regexp "^Bad signature from user.*$"
  "*Regular expression not matching a PGP signature validation message."
  :group 'x-pgp-sig
  :type 'regexp)

(defcustom x-pgp-sig-expected-re "Key ID \\(\\S +\\) not found"
  "*"
  :group 'x-pgp-sig
  :type 'regexp)

(defcustom x-pgp-sig-good-passphrase-regexp-for-decryption
  "\\.+Pass phrase appears good\\. \\."
  "*Regular expression matching PGP decryption \"good\" message."
  :group 'x-pgp-sig
  :type 'regexp)

(defcustom x-pgp-sig-verify-method-alist
  '(
    (gnus-summary-mode
     . (x-pgp-sig-gnus-get-article
	x-pgp-sig-find-gnus-article-buffer
	x-pgp-sig-gnus-unwind-after-verify))
    (gnus-Subject-mode
     . (x-pgp-sig-gnus-get-article
	x-pgp-sig-find-gnus-article-buffer
	x-pgp-sig-gnus-unwind-after-verify))
    (mh-folder-mode
     . (x-pgp-sig-mh-get-article
	x-pgp-sig-msg-to-verify-buffer))
    (mew-summary-mode
     . (x-pgp-sig-mew-get-article
	x-pgp-sig-msg-to-verify-buffer))
    (mew-virtual-mode
     . (x-pgp-sig-mew-get-article
	x-pgp-sig-msg-to-verify-buffer))
    (cmail-summary-mode
     . (x-pgp-sig-cmail-get-article
	x-pgp-sig-msg-to-verify-buffer))
    )
  "*Alist for verifying signature.
Each cells are:
'(MAJOR-MODE . (PREPARE-FUNCTION BUFFERorFUNCTION POST-FUNCTION))"
  :group 'x-pgp-sig
  :type 'alist)

(defcustom x-pgp-sig-debug nil
  "*"
  :group 'x-pgp-sig
  :type 'boolean)


;;; Internal variables
(defvar x-pgp-sig-no-conversion
  (cond (; XEmacs
	 running-xemacs (coding-system-name (get-coding-system nil)))
	(; Mule 2.3
	 (boundp 'MULE) '*noconv*)
	(; Emacs 20.x, MULE 3.0
	 (fboundp 'check-coding-system)
	 (or (condition-case ()
		 (check-coding-system 'binary)
	       (coding-system-error nil))
	     (condition-case ()
		 (check-coding-system 'no-conversion)
	       (coding-system-error nil))
	     ))))

(defconst x-pgp-sig-field-name "X-PGP-Sig"
  "Field name of PGP signature.")

(defconst x-pgp-sig-signed-headers-field-name "X-Signed-Headers"
  "Field name of PGP signed fields.")

(defconst x-pgp-sig-msg-begin-line "-----BEGIN PGP MESSAGE-----"
  "Text for start of PGP message delimiter.")

(defconst x-pgp-sig-msg-end-line "-----END PGP MESSAGE-----"
  "Text for end of PGP message delimiter.")

(defconst x-pgp-sig-signed-msg-begin-line "-----BEGIN PGP SIGNED MESSAGE-----"
  "Text for start of PGP signed messages.")

(defconst x-pgp-sig-signed-begin-line "-----BEGIN PGP SIGNATURE-----"
  "Text for start of PGP signature.")

(defconst x-pgp-sig-signed-end-line "-----END PGP SIGNATURE-----"
  "Text for end of PGP signature.")

(defvar x-pgp-sig-working-buffer " *x-pgp-sig-working*")

(defvar x-pgp-sig-msg-to-verify-buffer nil)

(defvar x-pgp-sig-enable-sign)
(setq x-pgp-sig-enable-sign x-pgp-sig-always-sign)

(or (stringp x-pgp-sig-default-signer)
    (setq x-pgp-sig-default-signer
	  (format "%s <%s>" (user-full-name) x-pgp-sig-from)))

(defvar x-pgp-sig-current-signer)
(setq x-pgp-sig-current-signer x-pgp-sig-default-signer)

(defvar x-pgp-sig-passphrase-alist)
(setq x-pgp-sig-passphrase-alist nil)

(defvar x-pgp-sig-public-key-alist)
(setq x-pgp-sig-public-key-alist nil)

(defvar x-pgp-sig-secret-key-alist)
(setq x-pgp-sig-secret-key-alist nil)

(defconst x-pgp-sig-idea-passphrase
  (let ((i 256)
	p r)
    (while (> i 0)
      (decf i)
      (setq r (abs (random))
	    p
	    (concat
	     p
	     (char-to-string
	      (+ (/ (* 95 (- r (* (/ r 100) 100))) 100) 32)))))
    p))


;;; Emulations for each Emacs variants.

(and (boundp 'to-kanji-display)
     (setq to-ascii-display ?B
	   to-ascii-fileio ?B
	   to-ascii-process ?B
	   to-kanji-display ?B
	   to-kanji-fileio ?B
	   to-kanji-process ?B))

(mapcar
 (function
  (lambda (elt)
    (let ((func (car elt)))
      (or (fboundp func)
	  (fset func (symbol-function (cdr elt)))))))
 '(
   (buffer-substring-no-properties . buffer-substring)
   (defsubst . defun)
   (read-char-exclusive . read-char)
   ))

(or (fboundp 'buffer-disable-undo)
    (defun buffer-disable-undo (&optional buffer)
      "Make BUFFER stop keeping undo information."
      (buffer-flush-undo (or buffer (current-buffer))))
    )

(or (fboundp 'delete)
    ;;; Imported from emu-18.el
    (defun delete (elt list)
      "Delete by side effect any occurrences of ELT as a member of LIST.
The modified LIST is returned.  Comparison is done with `equal'.
If the first member of LIST is ELT, deleting it is not a side effect;
it is simply using a different list.
Therefore, write `(setq foo (delete element foo))'
to be sure of changing the value of `foo'."
      (if (equal elt (car list))
	  (cdr list)
	(let ((rest list)
	      (rrest (cdr list))
	      )
	  (while (and rrest (not (equal elt (car rrest))))
	    (setq rest rrest
		  rrest (cdr rrest))
	    )
	  (rplacd rest (cdr rrest))
	  list))))

(or (fboundp 'match-string)
    (defun match-string (num &optional string)
      "Return string of text matched by last search.
NUM specifies which parenthesized expression in the last regexp.
Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING."
      (if (match-beginning num)
	  (if string
	      (substring string (match-beginning num) (match-end num))
	    (buffer-substring (match-beginning num) (match-end num))))))

(or (fboundp 'member)
    (defun member (elt list)
      "Return non-nil if ELT is an element of LIST. Comparison done with EQUAL.
The value is actually the tail of LIST whose car is ELT."
      (while (and list (not (equal elt (car list))))
	(setq list (cdr list)))
      list))

(or (fboundp 'put-alist)
    ;;; Imported from tl-list.el.
    (defun put-alist (item value alist)
      "Modify ALIST to set VALUE to ITEM.
If there is a pair whose car is ITEM, replace its cdr by VALUE.
If there is not such pair, create new pair (ITEM . VALUE) and
return new alist whose car is the new pair and cdr is ALIST."
      (let ((pair (assoc item alist)))
	(if pair
	    (progn
	      (setcdr pair value)
	      alist)
	  (cons (cons item value) alist)
	  ))))


;;; Function for DOSish OSes (OS/2,M$Windoze-95|NT).

(defsubst x-pgp-sig-delcr ()
  (goto-char (point-min))
  (while (search-forward "\r\n" nil t)
    (replace-match "\n")))


;;; Functions.

(defsubst x-pgp-sig-buffer-live-p (buffer)
  "Say whether BUFFER is alive or not."
  (and buffer
       (get-buffer buffer)
       (buffer-name (get-buffer buffer))))

(defun x-pgp-sig-read-noecho (prompt &optional stars)
  "Read a single line of text from user without echoing, and return it."
  (let ((ans "")
	(c 0)
	(echo-keystrokes 0)
	(cursor-in-echo-area t)
	(log-message-max-size 0)
	message-log-max done msg truncate)
    (while (not done)
      (if (or (not stars) (string-equal "" ans))
	  (setq msg prompt)
	(setq msg (concat prompt (make-string (length ans) ?*)))
	(setq truncate
	      (1+ (- (length msg) (window-width (minibuffer-window)))))
	(and (> truncate 0)
	     (setq msg (concat "$" (substring msg (1+ truncate))))))
      (message msg)
      (setq c (read-char-exclusive))
      (cond ((= c ?\C-g)
	     (setq quit-flag t
		   done t))
	    ((or (= c ?\r) (= c ?\n) (= c ?\e))
	     (setq done t))
	    ((= c ?\C-u)
	     (setq ans ""))
	    ((and (/= c ?\b) (/= c ?\177))
	     (setq ans (concat ans (char-to-string c))))
	    ((> (length ans) 0)
	     (setq ans (substring ans 0 -1)))))
    (if quit-flag
	(prog1
	    (setq quit-flag nil)
	  (message "Quit")
	  (beep t))
      (message "")
      ans)))

(defsubst x-pgp-sig-insert-file-contents-no-conversion (filename &rest args)
  (let (; NEmacs
	kanji-flag
	; Mule
	(input-coding-system x-pgp-sig-no-conversion)
	; XEmacs (>=20.1b6)
	(coding-system-for-read x-pgp-sig-no-conversion))
    (apply 'insert-file-contents filename args)))

(defun x-pgp-sig-error (&rest args)
  "(x-pgp-sig-error FORMAT ARG ARG ... BUFFER)."
  (let* ((len (length args))
	 (buffer (nth (1- len) args)))
    (cond ((zerop len) (error ""))
	  ((and (> len 1) (x-pgp-sig-buffer-live-p buffer))
	   (and x-pgp-sig-debug
		(let ((errbuf (generate-new-buffer
			       (concat "*x-pgp-sig-error-("
				       (buffer-name buffer)
				       ")*")))
		      content point)
		  (save-excursion
		    (set-buffer buffer)
		    (setq content (buffer-substring-no-properties
				   (point-min)
				   (point-max))
			  point (point))
		    (set-buffer errbuf)
		    (buffer-disable-undo)
		    (insert content)
		    (goto-char point)
		    (set-buffer-modified-p nil)
		    (setq buffer-read-only t)
		    (cond ((fboundp 'set-buffer-file-coding-system)
			   (set-buffer-file-coding-system
			    x-pgp-sig-no-conversion))
			  ((fboundp 'set-file-coding-system)
			   (set-file-coding-system x-pgp-sig-no-conversion))
			  ((fboundp 'set-kanji-fileio-code)
			   (set-kanji-fileio-code nil))))
		  (display-buffer errbuf)))
	   (apply 'error (reverse (cdr (reverse args)))))
	  (t (apply 'error args)))))

(defsubst x-pgp-sig-narrow-to-header (&optional include-separator)
  (widen)
  (goto-char (point-min))
  (narrow-to-region
   (point)
   (or (and (re-search-forward
	     (format "^\\(%s\\)?\n"
		     (regexp-quote mail-header-separator))
	     nil t)
	    (if include-separator
		(match-end 0)
	      (match-beginning 0)))
       (point-max)))
  (goto-char (point-min)))

(defun x-pgp-sig-process-buffer (args &optional passphrase buffer)
  (let ((obuf (or (and (x-pgp-sig-buffer-live-p buffer) buffer)
		  (current-buffer)))
	(pgpbuf (get-buffer-create " *PGP*"))
	(process-environment process-environment)
	process-connection-type case-fold-search envs env str err)
    (mapcar
     (function (lambda (a) (or (member a args) (setq args (cons a args)))))
     '("+verbose=1" "+language=en" "+batchmode"))
    (save-excursion
      (set-buffer obuf)
      (setq str (buffer-string))
      (set-buffer pgpbuf)
      (buffer-disable-undo)
      (setq buffer-read-only nil)
      (erase-buffer)
      (setq case-fold-search nil)
      (if (fboundp 'setenv)
	  (setenv "PGPPASSFD" (and passphrase "0"))
	(setq envs process-environment)
	(while (setq env (pop envs))
	  (and (string-match "^PGPPASSFD=" env)
	       (setq process-environment (delete env process-environment))))
	(and passphrase
	     (setq process-environment
		   (cons "PGPPASSFD=0" process-environment))))
      (let* (; NEmacs
	     (default-kanji-process-code
	       (or x-pgp-sig-process-kanji-code 0))
	     program-kanji-code-alist
	     ;
	     (outcode (or x-pgp-sig-process-coding-system-for-output
			  x-pgp-sig-no-conversion))
	     ; Mule
	     (default-process-coding-system
	       (cons x-pgp-sig-no-conversion outcode))
	     call-process-hook
	     ; XEmacs (<= 20.1b6)
	     (file-coding-system outcode)
	     ; XEmacs (>= 20.1b7)
	     (buffer-file-coding-system outcode)
	     ; XEmacs (>= 20.2b1)
	     (coding-system-for-read x-pgp-sig-no-conversion)
	     (coding-system-for-write outcode)
	     )
	(and passphrase (insert passphrase "\n"))
	(insert str)
	(apply 'call-process-region (point-min) (point-max)
	       x-pgp-sig-pgp-program t pgpbuf nil args))
      ;;; for 'OS/2 & 'windoze-97 convert CRLF to LF
      (x-pgp-sig-delcr)
      ;;;
      (goto-char (point-min))
      (and (re-search-forward "\n\n+" nil t)
	   (looking-at "^\C-gError:[\t ]*\\(.*\\)$")
	   (setq err (match-string 1)))
      (setq str (buffer-string))
      (set-buffer obuf)
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert str))
    (if err
	(list (format "\"%s\"" err) pgpbuf)
      (kill-buffer pgpbuf)
      nil)))

(defun x-pgp-sig-idea-encrypt-string (str)
  "Encrypt string using IDEA method."
  (and (stringp str)
       (let ((buffer (get-buffer-create x-pgp-sig-working-buffer))
	     beg result)
	 (unwind-protect
	     (save-excursion
	       (set-buffer buffer)
	       (buffer-disable-undo)
	       (setq buffer-read-only nil)
	       (erase-buffer)
	       (insert str)
	       (setq str nil)
	       (setq result (x-pgp-sig-process-buffer
			     '("-fact") x-pgp-sig-idea-passphrase))
	       (and result (apply 'x-pgp-sig-error "!! %s" result))
	       (goto-char (point-min))
	       (or (re-search-forward
		    (format "%s$" (regexp-quote x-pgp-sig-msg-begin-line))
		    nil t)
		   (x-pgp-sig-error
		    "!! Regexp \"%s\$\" not found"
		    (regexp-quote x-pgp-sig-msg-begin-line)
		    buffer))
	       (or (search-forward "\n\n" nil t)
		   (x-pgp-sig-error "!! String \"\\n\\n\" not found"
				    buffer))
	       (setq beg (point))
	       (or (re-search-forward
		    (format "%s$" (regexp-quote x-pgp-sig-msg-end-line))
		    nil t)
		   (x-pgp-sig-error
		    "!! Regexp \"%s\$\" not found"
		    (regexp-quote x-pgp-sig-msg-end-line)
		    buffer))
	       (setq result (buffer-substring beg (match-beginning 0))))
	   (kill-buffer buffer))
	 result)))

(defun x-pgp-sig-idea-decrypt-string (str)
  "Decrypt string using IDEA method."
  (and (stringp str)
       (let ((buffer (get-buffer-create x-pgp-sig-working-buffer))
	     result)
	 (unwind-protect
	     (save-excursion
	       (set-buffer buffer)
	       (buffer-disable-undo)
	       (setq buffer-read-only nil)
	       (erase-buffer)
	       (insert x-pgp-sig-msg-begin-line "\n\n" str)
	       (setq str nil)
	       (or (eolp) (insert "\n"))
	       (insert x-pgp-sig-msg-end-line "\n")
	       (setq result
		     (x-pgp-sig-process-buffer
		      '("-f") x-pgp-sig-idea-passphrase))
	       (and result (apply 'x-pgp-sig-error "!! %s" result))
	       (goto-char (point-min))
	       (or (re-search-forward
		    x-pgp-sig-good-passphrase-regexp-for-decryption nil t)
		   (x-pgp-sig-error
		    "!! Regexp \"%s\" not found"
		    (regexp-quote
		     x-pgp-sig-good-passphrase-regexp-for-decryption)
		    buffer))
	       (setq result (buffer-substring (point) (point-max))))
	   (kill-buffer buffer))
	 result)))

(defun x-pgp-sig-lookup-keys (&optional regexp case pub)
  "Returns UserID/KeyID alist."
  (let ((case-fold-search t)
	(keyring (if pub x-pgp-sig-pubring x-pgp-sig-secring))
	(re (format "^%s[\t ]+[0-9]+/\\([0-9a-f]+\\)[\t ]+[^\n\t ]+"
		    (if pub "pub" "sec")))
	process-connection-type
	buffer id result cell match)
    (or (file-exists-p keyring)
	(error "!! File \"%s\" not found" keyring))
    (unwind-protect
	(save-excursion
	  (setq buffer (set-buffer
			(get-buffer-create x-pgp-sig-working-buffer)))
	  (buffer-disable-undo)
	  (setq case-fold-search t
		buffer-read-only nil)
	  (erase-buffer)
	  (call-process
	   x-pgp-sig-pgp-program nil t nil
	   ;;; Suggested by Kazuyoshi Michishita <mit@unya.rim.or.jp>
	   "+verbose=1"
	   ;;;
	   "+language=en" "-kv" keyring)
	  ;;; for 'windoze-97
	  (x-pgp-sig-delcr)
	  ;;;
	  (goto-char (point-min))
	  (when (re-search-forward
		 (format "^Key ring: '%s'\nType.+\n" (regexp-quote keyring))
		 nil t)
	    (while (re-search-forward re nil t)
	      (setq id (concat "0x" (match-string 1)))
	      (save-restriction
		(narrow-to-region
		 (match-end 0)
		 (or (and (re-search-forward "^[^\n\t ]" nil t)
			  (match-beginning 0))
		     (point-max)))
		(goto-char (point-min))
		(while (re-search-forward "[\t ]+\\(.+\\)$" nil t)
		  (setq result
			(append result (list (cons (match-string 1) id)))))
		(goto-char (point-max)))
	      )))
      (kill-buffer buffer))
    (if (stringp regexp)
	(progn
	  (setq case-fold-search case)
	  (while (setq cell (pop result))
	    (and (string-match regexp (car cell))
		 (setq match (append match (list cell)))))
	  match)
      result)))

(defun x-pgp-sig-build-message-to-sign ()
  "Extract fields and body to sign in current buffer."
  (let ((case-fold-search t)
	(sign-fields (copy-sequence x-pgp-sig-sign-fields))
	(adjust x-pgp-sig-adjust-sign-fields-alist)
	(fields "")
	adj adj1 sub add existing-fields field re body buffer)
    (save-excursion
      (save-restriction
	(x-pgp-sig-narrow-to-header t)
	(while (setq adj (pop adjust))
	  (setq adj1 (pop adj))
	  (goto-char (point-min))
	  (when (cond ((stringp adj1) (re-search-forward adj1 nil t))
		      ((fboundp adj1) (funcall adj1)))
	    (setq sub (cdr (assq 'sub adj))
		  add (cdr (assq 'add adj)))
	    (while sub
	      (setq sign-fields (delete (car sub) sign-fields))
	      (setq sub (cdr sub)))
	    (while add
	      (setq sign-fields (append sign-fields (list (car add)))
		    add (cdr add)))))
	(while (setq field (pop sign-fields))
	  (setq re (format "^%s:" (regexp-quote field)))
	  (goto-char (point-min))
	  (and (re-search-forward re nil t)
	       (or (re-search-forward re nil t)
		   (setq fields (concat fields
					(buffer-substring-no-properties
					 (match-beginning 0)
					 (or (and (re-search-forward
						   "^[^\n\t ]\\|^$" nil t)
						  (match-beginning 0))
					     (point-max))))
			 existing-fields (append existing-fields
						 (list field))))))
	(or existing-fields
	    (error "!! No fields to sign"))
	(goto-char (point-max))
	(widen)
	(setq body (buffer-substring-no-properties (point) (point-max)))
	(setq buffer
	      (set-buffer (get-buffer-create " *x-pgp-sig-msg-to-verify*")))
	(buffer-disable-undo)
	(setq buffer-read-only nil)
	(erase-buffer)
	(insert body)
	(goto-char (point-min))
	(insert x-pgp-sig-signed-headers-field-name ": "
		(mapconcat 'identity existing-fields ",") "\n"
		fields "\n")
	(cons existing-fields buffer)))))

(defun x-pgp-sig-create-x-pgp-signature ()
  "To sign the message and create field body."
  (let ((case-fold-search t)
	(msgs (x-pgp-sig-build-message-to-sign))
	(passphrase (x-pgp-sig-idea-decrypt-string
		     (cdr (assoc x-pgp-sig-current-signer
				 x-pgp-sig-passphrase-alist))))
	buffer result beg version)
    (or passphrase
	(setq passphrase (x-pgp-sig-set-signer)))
    (unwind-protect
	(save-excursion
	  (set-buffer (setq buffer (cdr msgs)))
	  (setq case-fold-search t)
	  (run-hooks 'x-pgp-sig-prepare-sign-hook)
	  (setq result
		(x-pgp-sig-process-buffer
		 (list "-fastu" x-pgp-sig-current-signer) passphrase))
	  (and result (apply 'x-pgp-sig-error "!! %s" result))
	  (run-hooks 'x-pgp-sig-post-sign-hook)
	  (goto-char (point-min))
	  (or (re-search-forward
	       (format "^%s\n" (regexp-quote x-pgp-sig-signed-begin-line))
	       nil t)
	      (x-pgp-sig-error
	       "!! Regexp \"%s\\n\" not found"
	       (regexp-quote x-pgp-sig-signed-begin-line)
	       buffer))
	  (setq beg (point))
	  (or (re-search-forward
	       (format "^%s$" (regexp-quote x-pgp-sig-signed-end-line))
	       nil t)
	      (x-pgp-sig-error
	       "!! Regexp \"%s\$\" not found"
	       (regexp-quote x-pgp-sig-signed-end-line)
	       buffer))
	  (narrow-to-region beg (match-beginning 0))
	  (goto-char (point-min))
	  (setq version
		(or
		 (and (re-search-forward "^version:[\t ]*\\([^\n\t ]+\\)"
					 nil t)
		      (buffer-substring (match-beginning 1) (match-end 1)))
		 "0.0"))
	  (or (re-search-forward "\n\n\\([^\n\t ]\\)" nil t)
	      (x-pgp-sig-error
	       "!! Regexp \"\\n\\n\\\\(\[\^\\n\\t \]\\\\)\" not found"
	       buffer))
	  (delete-region (point-min) (match-beginning 1))
	  (goto-char (point-min))
	  (insert version " "
		  (mapconcat 'identity (car msgs) ",")
		  "\n")
	  (while (not (eobp))
	    (insert "\t")
	    (end-of-line)
	    (forward-char 1))
	  (setq msgs (buffer-substring (point-min) (1- (point-max)))))
      (kill-buffer buffer))
    msgs))

(defun x-pgp-sig-build-message-to-verify (&optional buffer silent)
  "Extract header and body to verify in specified buffer."
  (if (and buffer (not (x-pgp-sig-buffer-live-p buffer)))
      (format "!! Buffer %s not exists"
	      (prin1-to-string buffer))
    (and buffer (set-buffer (get-buffer buffer)))
    (let ((case-fold-search t)
	  lines version beg end fields fieldsx field
	  pgpsig pgphead body clines ldiff vbuff)
      (save-restriction
	(x-pgp-sig-narrow-to-header t)
	(and (re-search-forward "^lines:[\t ]+\\([0-9]+\\)" nil t)
	     (setq lines (string-to-int (match-string 1))))
	(goto-char (point-min))
	(if (re-search-forward
	     (format "^%s:%s%s"
		     (regexp-quote x-pgp-sig-field-name)
		     "[\t ]+\\([^\n\t ,]+\\)[\t ]+\\([^\n\t ]+\\)"
		     "\\(\n\\([\t ]+.+\n\\)+\\|\\([\t ]+[^\n\t ]+\\)+\n\\)"
		     ) nil t)
	    (progn
	      (setq version (match-string 1)
		    beg (match-end 2)
		    end (match-end 0))
	      (goto-char (match-beginning 2))
	      (while (re-search-forward "[^\n\t ,]+" beg t)
		(setq fields (append fields (list (match-string 0)))))
	      (setq fieldsx fields)
	      (goto-char beg)
	      (while (re-search-forward "[\t ]+\\([^\n\t ]+\\)" end t)
		(setq pgpsig (concat pgpsig (match-string 1) "\n")))
	      (while (setq field (pop fields))
		(goto-char (point-min))
		(setq pgphead
		      (concat pgphead field
			      (or (and (re-search-forward
					(format "^\\(%s\\):\\([\t ]+.+\n\\)+"
						(regexp-quote field))
					nil t)
				       (buffer-substring-no-properties
					(match-end 1) (match-end 0)))
				  ": \n"))))
	      (goto-char (point-max))
	      (widen)
	      (setq body (buffer-substring-no-properties (point) (point-max))
		    clines (count-lines (point) (point-max))
		    ldiff (or (and lines (- lines clines)) 0)
		    vbuff (set-buffer
			   (get-buffer-create " *x-pgp-sig-msg-to-verify*")))
	      (buffer-disable-undo)
	      (setq buffer-read-only nil)
	      (erase-buffer)
	      (insert body)
	      (or (eolp) (insert "\n"))
	      (goto-char (point-min))
	      (while (re-search-forward "^-" nil t)
		(replace-match "- -"))
	      (goto-char (point-min))
	      (insert x-pgp-sig-signed-msg-begin-line "\n"
		      "\n"
		      x-pgp-sig-signed-headers-field-name ": "
		      (mapconcat 'identity fieldsx ",") "\n"
		      pgphead
		      "\n")
	      (goto-char (point-max))
	      (insert "\n"
		      x-pgp-sig-signed-begin-line "\n"
		      "Version: " version "\n"
		      "\n"
		      pgpsig
		      x-pgp-sig-signed-end-line "\n")
	      (cons ldiff vbuff))
	  (if silent
	      ""
	    (format "!! Field \"%s\" not found"
		    x-pgp-sig-field-name)))))))

(defun x-pgp-sig-put-field (field-name field-body)
  "Insert field without dragging faces."
  (let ((case-fold-search t)
	(re (format "^%s:" (regexp-quote field-name)))
	(last (format "^\\(%s\\):"
		      (mapconcat 'identity x-pgp-sig-last-fields "\\|"))))
    (save-excursion
      (goto-char (point-min))
      (save-restriction
	(narrow-to-region
	 (point)
	 (or (and (re-search-forward
		   (format "^\\(%s\\)?\n" (regexp-quote mail-header-separator))
		   nil t)
		  (match-beginning 0))
	     (point-max)))
	(goto-char (point-min))
	(while (re-search-forward re nil t)
	  (delete-region
	   (match-beginning 0)
	   (or (and (re-search-forward "^[^\n\t ]\\|^$" nil t)
		    (match-beginning 0))
	       (point-max))))
	(goto-char (point-min))
	(if (re-search-forward last nil t)
	    (goto-char (match-beginning 0))
	  (goto-char (point-max)))
	(if (bobp)
	    (insert field-name ": " field-body "\n")
	  (backward-char 1)
	  (insert "\n" field-name ": " field-body))))))


;;; Commands and functions to be autoloaded.

;;;###autoload
(defun x-pgp-sig-sign (&optional force)
  "Add X-PGP signature."
  (let ((case-fold-search t)
	end clength signature)
    (save-excursion
      (save-restriction
	(widen)
	(goto-char (point-min))
	(when (re-search-forward
	       (format "^\\(%s\\)?\n" (regexp-quote mail-header-separator))
	       nil t)
	  (setq end (match-beginning 0)
		clength (- (point-max) (match-end 0)))
	  (goto-char (point-min))
	  (and
	   (not (zerop clength))
	   (or force
	       x-pgp-sig-always-sign
	       (and
		(not (re-search-forward
		      "^control:[\t ]*cancel\\|^resent-\\(to\\|cc\\):"
		      end t))
		(or x-pgp-sig-enable-sign
		    (and (fboundp x-pgp-sig-add-signature-judging-function)
			 (funcall x-pgp-sig-add-signature-judging-function)))))
	   (progn
	     (and x-pgp-sig-approved-field-body
		  (re-search-forward "^control:" end t)
		  (x-pgp-sig-put-field "Approved"
				       x-pgp-sig-approved-field-body))
	     (and x-pgp-sig-x-info-field-body
		  (x-pgp-sig-put-field "X-Info"
				       x-pgp-sig-x-info-field-body))
	     (when
		 (setq signature
		       (condition-case err
			   (progn
			     (message "Signing the message...")
			     (prog1
				 (x-pgp-sig-create-x-pgp-signature)
			       (message "Signing the message...done")))
			 (error
			  (progn
			    (message (if (and (listp err)
					      (stringp (car (cdr err))))
					 (car (cdr err))
				       (prin1-to-string err)))
			    nil))))
	       (x-pgp-sig-put-field x-pgp-sig-field-name signature)
	       (run-hooks 'x-pgp-sig-sign-hook))))))))
  (setq x-pgp-sig-enable-sign nil))

;;;###autoload
(defun x-pgp-sig-set-signer (&optional signer passphrase)
  "Set PGP signer and passphrase."
  (interactive)
  (let ((signer (or signer
		    (if (stringp x-pgp-sig-current-signer)
			x-pgp-sig-current-signer
		      (user-login-name))))
	(keys (or x-pgp-sig-secret-key-alist
		  (setq x-pgp-sig-secret-key-alist (x-pgp-sig-lookup-keys))
		  (error "!! No keys in \"%s\"" x-pgp-sig-secring)))
	(interactive (interactive-p))
	case-fold-search orig id key result)
    (setq orig signer)
    (and (string-match "^[^\n\t <>]+$" signer)
	 (setq case-fold-search t))
    (while (setq key (pop keys))
      (and (string-match signer (car key))
	   (setq id (car key)
		 keys nil)))
    (setq signer
	  (if interactive
	      (completing-read "Enter the key's user ID: "
			       x-pgp-sig-secret-key-alist nil t id)
	    (if id
		id
	      (error "!! Secret key for \"%s\" not found" orig))))
    (setq key (cdr (assoc signer x-pgp-sig-secret-key-alist)))
    (when (or interactive
	      (not (or passphrase (assoc signer x-pgp-sig-passphrase-alist))))
      (setq interactive t)
      (setq passphrase (x-pgp-sig-read-noecho
			(format "PGP passphrase for %s (%s): " signer key)
			t))
      (and (zerop (length passphrase)) (error "!! Empty passphrase")))
    (and passphrase
	 (when interactive
	   (message "Confirming passphrase...")
	   (let ((buffer (get-buffer-create x-pgp-sig-working-buffer)))
	     (unwind-protect
		 (save-excursion
		   (set-buffer buffer)
		   (buffer-disable-undo)
		   (setq buffer-read-only nil)
		   (erase-buffer)
		   (setq result
			 (x-pgp-sig-process-buffer
			  (list "-fastu" signer) passphrase))
		   (if result
		       (apply 'x-pgp-sig-error "!! %s" result)
		     (message "Confirming passphrase...done")))
	       (kill-buffer buffer))))
	 (setq x-pgp-sig-passphrase-alist
	       (put-alist signer
			  (progn (message "Encrypting passphrase...")
				 (x-pgp-sig-idea-encrypt-string passphrase))
			  x-pgp-sig-passphrase-alist)))
    (setq x-pgp-sig-current-signer signer)
    (prog1
	(or passphrase
	    (progn
	      (message "Decrypting passphrase...")
	      (x-pgp-sig-idea-decrypt-string
	       (cdr (assoc signer x-pgp-sig-passphrase-alist)))))
      (message "The signer and the passphrase have been registered."))))

;;;###autoload
(defun x-pgp-sig-set-signer-maybe ()
  "Set PGP signer and passphrase if necessary."
  (interactive)
  (or (stringp x-pgp-sig-default-signer)
      (setq x-pgp-sig-default-signer
	    (format "%s <%s>" (user-full-name) x-pgp-sig-from)))
  (and (or
	(not (string-equal x-pgp-sig-current-signer x-pgp-sig-default-signer))
	(not (assoc x-pgp-sig-current-signer x-pgp-sig-passphrase-alist)))
       (condition-case err
	   (call-interactively 'x-pgp-sig-set-signer)
	 (error (message "%s" (if (and (listp err) (stringp (car (cdr err))))
				  (car (cdr err))
				(prin1-to-string err)))))))

;;;###autoload
(defun x-pgp-sig-switch-signing (&optional switch)
  "Judging whether to add X-PGP signature."
  (interactive
   (list (y-or-n-p
	  (format "Do you want to add %s field? " x-pgp-sig-field-name))))
  (let ((msg "This message will %sbe signed."))
    (if (setq x-pgp-sig-enable-sign (not (not switch)))
	(progn
	  (and (boundp 'mime-editor/pgp-processing)
	       (eq mime-editor/pgp-processing 'sign)
	       (setq mime-editor/pgp-processing nil))
	  (setq msg (format msg "")))
      (setq msg (format msg "not ")))
    (and (interactive-p) (message msg))))

;;;###autoload
(defun x-pgp-sig-verify ()
  "Verify X-PGP signature in current message."
  (interactive)
  (let ((method (assq major-mode x-pgp-sig-verify-method-alist))
	(delete x-pgp-sig-delete-last-empty-line-while-verifying)
	(re (format "\\(\n+\\)%s" (regexp-quote x-pgp-sig-signed-begin-line)))
	(count 0)
	prepare buffer post vbuff vbuf2 ldiff plus newlines good bad pgp-id)
    (and method
	 (setq method (cdr method)
	       prepare (car method)
	       buffer (car (cdr method))
	       post (car (cdr (cdr method)))))
    (and (interactive-p) (message "Verifying signature..."))
    (and prepare (funcall prepare))
    (setq buffer (condition-case ()
		     (if (fboundp buffer)
			 (funcall buffer)
		       (symbol-value buffer))
		   (error nil)))
    (or (x-pgp-sig-buffer-live-p buffer)
	(setq buffer (current-buffer)))
    (save-excursion
      (setq vbuff (x-pgp-sig-build-message-to-verify
		   buffer (not (interactive-p))))
      (unless (stringp vbuff)
	(setq ldiff (car vbuff)
	      plus (if (>= ldiff 0) "+" "")
	      vbuff (set-buffer (cdr vbuff)))
	(if (zerop ldiff)
	    (message "Verifying signature...")
	  (message "(%s%d/0) Verifying signature..." plus ldiff))
	(goto-char (point-min))
	(setq newlines
	      (or (and (re-search-forward re nil t)
		       (length (match-string 1)))
		  0))
	(setq vbuf2
	      (set-buffer
	       (get-buffer-create " *x-pgp-sig-msg-to-verify-2*")))
	(buffer-disable-undo)
	(setq buffer-read-only nil)
	(while (or delete (not (or good bad)))
	  (erase-buffer)
	  (insert-buffer-substring vbuff)
	  (and delete
	       (> count 0)
	       (> newlines (1+ count))
	       (goto-char (point-min))
	       (re-search-forward re nil t)
	       (goto-char (match-beginning 0))
	       (message "(%s%d/-%d) Verifying signature..."
			plus ldiff count)
	       (delete-char count))
	  (or (> newlines (+ 2 count))
	      (setq delete nil))
	  (x-pgp-sig-process-buffer '("-f") nil vbuf2)
	  (incf count)
	  (goto-char (point-min))
	  (and (re-search-forward x-pgp-sig-good-signature-regexp nil t)
	       (setq good (match-string 0)
		     delete nil))
	  (goto-char (point-min))
	  (if (re-search-forward x-pgp-sig-expected-re nil t)
	      (progn
		(setq pgp-id (concat "0x"
				     (buffer-substring-no-properties
				      (match-beginning 1) (match-end 1))))
		(or
		 (and (or (boundp 'mc-pgp-always-fetch)
			  (condition-case ()
			      (load-library "mc-pgp"))
			  (error nil))
		      (not (eq mc-pgp-always-fetch 'never))
		      (or mc-pgp-always-fetch
			  (y-or-n-p
			   (format
			    "Key %s not found; attempt to fetch? "
			    pgp-id)))
		      (mc-pgp-fetch-key (cons pgp-id nil))
		      (setq count 0))
		 (setq bad (format "!! Key %s not found" pgp-id))))
	    (goto-char (point-min))
	    (setq bad (concat "!! "
			      (or (and (re-search-forward
					x-pgp-sig-bad-signature-regexp
					nil t)
				       (match-string 0))
				  "Bad signature")
			      ""))))))
    (if (stringp vbuff)
	(message vbuff)
      (and vbuff (kill-buffer vbuff))
      (and vbuf2 (kill-buffer vbuf2))
      (and post (funcall post))
      (and (x-pgp-sig-buffer-live-p x-pgp-sig-msg-to-verify-buffer)
	   (kill-buffer x-pgp-sig-msg-to-verify-buffer))
      (decf count)
      (message "%s%s"
	       (if (and (zerop ldiff) (zerop count))
		   ""
		 (format "(%s%d/-%d) " plus ldiff count))
	       (or good bad)))))

(defun x-pgp-sig-add-signature-y-or-n-p ()
  (y-or-n-p "Do you want to add X-PGP-Sig field? "))


;;; GNUS, Gnus
(defun x-pgp-sig-gnus-get-article ()
  (and (boundp 'gnus-article-display-hook)
       (memq 'x-pgp-sig-verify-for-gnus gnus-article-display-hook)
       (setq gnus-article-display-hook
	     (append (delq 'x-pgp-sig-verify-for-gnus
			   gnus-article-display-hook)
		     '(x-pgp-sig-verify-for-gnus))))
  (or (and (boundp 'gnus-original-article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-original-article-buffer)
	   (eq gnus-current-article (gnus-summary-article-number)))
      (cond ((eq 'gnus-summary-mode major-mode)
	     (if (boundp 'gnus-original-article-buffer)
		 ; >= v5.2
		 (gnus-summary-show-article)
	       ; <= v5.1
	       (let ((gnus-show-all-headers t)
		     gnus-show-mime)
		 (gnus-summary-show-article))))
	    ((eq 'gnus-Subject-mode major-mode)
	     ; <= 3.14.4
	     (let ((gnus-show-all-headers t))
	       (gnus-Subject-show-article))))))

(defun x-pgp-sig-find-gnus-article-buffer ()
  "Find article buffer may includes X-PGP-Sig field."
  (or (and (boundp 'gnus-original-article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-original-article-buffer))
      (and (boundp 'gnus-article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-article-buffer))
      (and (boundp 'gnus-Article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-Article-buffer))))

(defun x-pgp-sig-gnus-unwind-after-verify ()
  (or (boundp 'gnus-original-article-buffer)
      (cond ((eq 'gnus-summary-mode major-mode)
	     (gnus-summary-show-article))
	    ((eq 'gnus-Subject-mode major-mode)
	     (gnus-Subject-show-article)))))

(defun x-pgp-sig-verify-for-gnus ()
  (and gnus-show-mime
       (save-excursion
	 (set-buffer gnus-summary-buffer)
	 (x-pgp-sig-verify))))

;;; MH-E
(defun x-pgp-sig-mh-get-article ()
  "Get current message in article buffer."
  (let ((file (concat mh-folder-filename (mh-get-msg-num t))))
    (save-excursion
      (setq x-pgp-sig-msg-to-verify-buffer
	    (set-buffer (get-buffer-create " *x-pgp-sig-tmp-article*")))
      (buffer-disable-undo)
      (setq buffer-read-only nil)
      (erase-buffer)
      (x-pgp-sig-insert-file-contents-no-conversion file)
      (x-pgp-sig-delcr))))

;;; Mew
(defun x-pgp-sig-mew-get-article ()
  "Get current message in article buffer."
  (cond
   ((mew-summary-part-number)
    (re-search-backward mew-summary-message-regex nil t nil)
    (mew-summary-display))
   )
  (let ((file
	 (if (fboundp 'mew-expand-folder)
	     (mew-expand-folder (mew-summary-folder-name)     ;; Mew 1.78 -
				(mew-summary-message-number))
	   (mew-expand-file-name (mew-summary-message-number) ;; - Mew 1.76
				 (mew-summary-folder-name)))))
    (save-excursion
      (setq x-pgp-sig-msg-to-verify-buffer
	    (set-buffer (get-buffer-create " *x-pgp-sig-tmp-article*")))
      (buffer-disable-undo)
      (setq buffer-read-only nil)
      (erase-buffer)
      (x-pgp-sig-insert-file-contents-no-conversion file)
      (x-pgp-sig-delcr))))

(defun x-pgp-sig-mew-encode-message-header-with-semi ()
  "Encode the message header to MIME format with SEMI."
  (require 'eword-encode)
  (eword-encode-header))

(defun x-pgp-sig-mew-encode-message-header-with-tm ()
  "Encode the message header to MIME format with tm."
  (require 'tm-ew-e)
  (mime/encode-message-header))

(defun x-pgp-sig-mew-encode-message-header-with-toMime ()
  "Encode the message header to MIME format with toMime."
  (save-excursion
    (save-restriction
      (x-pgp-sig-narrow-to-header)
      (let (
	    ; Mule-2.x
	    (input-coding-system mew-cs-noconv)
	    (output-coding-system mew-cs-mime-trans)
	    ; XEmacs
	    (coding-system-for-read mew-cs-noconv)
	    (coding-system-for-write mew-cs-mime-trans)
	    )
	(apply 'call-process-region (point-min) (point-max)
	       x-pgp-sig-toMime-program t t nil
	       x-pgp-sig-toMime-options)))))

(defun x-pgp-sig-mew-encode-message ()
  "Encode the message to the network code."
  (save-excursion
    (save-restriction
      (widen)
      (let ((inhibit-read-only t))
	(put-text-property (point-min) (point-max) 'read-only nil))
      (make-local-variable 'mail-header-separator)
      (setq mail-header-separator mew-header-separator
	    buffer-read-only nil)
      (and x-pgp-sig-mew-encode-message-header-function
	   (funcall x-pgp-sig-mew-encode-message-header-function))
      (goto-char (point-min))
      (and (re-search-forward "\\cj\\|\\ch\\|\\ck" nil t)
	   (mew-cs-encode-region (point-min) (point-max) mew-cs-mime-trans)))))

(defun x-pgp-sig-mew-encode-coding-and-sign (&optional force)
  "Encode the message to the network code and sign."
  (let ((inhibit-read-only t))
    (put-text-property (point-min) (point-max) 'read-only nil))
  (mew-cs-encode-region (point-min) (point-max) mew-cs-mime-trans)
  (x-pgp-sig-sign force))

;;; cmail
(defun x-pgp-sig-cmail-get-article ()
  "Get current message in article buffer."
  (let ((page (cmail-get-page-number-from-summary))
	(folder (get-buffer (cmail-folder-buffer cmail-current-folder)))
	beg end)
    (save-excursion
      (set-buffer folder)
      (cmail-n-page page)
      (setq beg (point))
      (cmail-n-page (1+ page))
      (forward-line -1)
      (setq end (point))
      (setq x-pgp-sig-msg-to-verify-buffer
	    (set-buffer (get-buffer-create " *x-pgp-sig-tmp-article*")))
      (buffer-disable-undo)
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert-buffer-substring folder beg end))))


(provide 'x-pgp-sig)
(run-hooks 'x-pgp-sig-load-hook)
;;; x-pgp-sig.el ends here.
